﻿from typing import List
import time
import argparse

from PyQt5 import QtCore
from PyQt5.QtCore import Qt, QThread
from PyQt5.QtWidgets import QApplication

from .. import config
from ..product.mode import Frequency, Voltage, Current
from ..oscilloscope.oscilloscopemeasurer import OscilloscopeMeasurer
from ..gpio.gpioreader import GpioState, GpioReader
from ..product.mode import CurveState
from ..screenshot import ExternalMediaState, ScreenshotFileManager
from .mainwindowbytype import main_window_by_type
from .._version import version as main_version
from PyQt5.QtCore import QTranslator

__all__ = ["Application"]


class Application(QApplication):
    __start_operations = QtCore.pyqtSignal()
    __stop_operations = QtCore.pyqtSignal()

    __measure_required = QtCore.pyqtSignal()
    __freeze_reference_curve_changed = QtCore.pyqtSignal(bool)
    __freeze_test_curve_changed = QtCore.pyqtSignal(bool)

    def __init__(self, args):
        super().__init__(args)
        self.__english = False
        self.setApplicationDisplayName(main_version)
        self.__reference_oscilloscope_path = None
        self.__test_oscilloscope_path = None

        self.__main_window_type = None

        self.__screenshot_external_media_path = None
        self.__screenshot_external_media_only = False
        self.__screenshots_show_save_as_dialog = True

        self.__screenshot_gpio = None
        self.__screenshot_gpio_inverted = False

        self.__freeze_gpio = None
        self.__freeze_gpio_inverted = False

        self.__hide_cursor = False

        self.__debug = False
        self.__enable_sound_support = True  # If False, sound will be disabled at all overriding value in config
        self.__enable_score = True  # If False, score calculation will be disabled overriding value in config

        self.__parse_args(args)

        if self.__english:  # translate language if flag --en is true
            translator = QTranslator()
            translator.load("eyepoint_s1/gui/super_translate_en")  # super_translate_en.qm
            if not self.installTranslator(translator):
                print("Can not install translation!")

        self.__lightweight_workers_thread = QThread(parent=self)
        self.__lightweight_workers_thread.start()

        self.__lightweight_workers = []

        self.__screenshot_file_manager = None
        self.__init_screenshot_manager()

        self.__main_window = None
        self.__init_main_window()

        self.__screenshot_gpio_reader = None
        self.__freeze_gpio_reader = None
        self.__init_gpio_readers()

        self.__measurer_thread = QThread(parent=self)
        self.__measurer_thread.start()

        self.__measurer = OscilloscopeMeasurer(
            reference_oscilloscope_path=self.__reference_oscilloscope_path,
            test_oscilloscope_path=self.__test_oscilloscope_path,

            frequency=config.frequency,
            voltage=config.voltage,
            current=config.current
        )
        self.__measurer.moveToThread(self.__measurer_thread)
        self.__start_operations.connect(self.__measurer.start)
        self.__stop_operations.connect(self.__measurer.stop)

        # For FPS
        self.__previous_frame_time = time.monotonic()

        # Signals
        self.__measure_required.connect(self.__measurer.async_measure)
        self.__freeze_reference_curve_changed.connect(self.__measurer.set_freeze_reference_curve)
        self.__freeze_test_curve_changed.connect(self.__measurer.set_freeze_test_curve)

        self.__main_window.frequency_changed.connect(self.__measurer.set_frequency)
        self.__main_window.voltage_changed.connect(self.__measurer.set_voltage)
        self.__main_window.current_changed.connect(self.__measurer.set_current)

        self.__main_window.reference_curve_state_changed.connect(self.__on_reference_curve_state_changed)
        self.__main_window.test_curve_state_changed.connect(self.__on_test_curve_state_changed)
        self.__main_window.external_freeze_state_changed.connect(self.__on_external_freeze_state_changed)

        self.__measurer.devices_ready.connect(self.__on_devices_ready)
        self.__measurer.devices_not_correct.connect(self.__on_devices_correct)
        self.__measurer.curves_ready.connect(self.__on_curves_ready)
        self.__main_window.auto_calibration.connect(self.__measurer.calibrate_oscilloscopes)

    @QtCore.pyqtSlot(CurveState)
    def __on_reference_curve_state_changed(self, reference_curve_state: CurveState) -> None:
        self.__freeze_reference_curve_changed.emit(reference_curve_state == CurveState.FROZEN)

    @QtCore.pyqtSlot(CurveState)
    def __on_test_curve_state_changed(self, test_curve_state: CurveState) -> None:
        self.__freeze_test_curve_changed.emit(test_curve_state == CurveState.FROZEN)

    @QtCore.pyqtSlot(CurveState)
    def __on_external_freeze_state_changed(self, external_curve_state: CurveState) -> None:
        self.__freeze_reference_curve_changed.emit(self.__main_window.reference_curve_state == CurveState.FROZEN)
        self.__freeze_test_curve_changed.emit(self.__main_window.test_curve_state == CurveState.FROZEN)

    def __add_workers(self, *workers) -> None:
        for worker in workers:
            worker.moveToThread(self.__lightweight_workers_thread)
            self.__start_operations.connect(worker.start)
            self.__stop_operations.connect(worker.stop)

        self.__lightweight_workers.extend(workers)

    def start_app(self) -> int:
        if config.maximize_window:
            self.__main_window.showMaximized()
        else:
            self.__main_window.show()

        self.__start_operations.emit()
        self.__measure_required.emit()  # Starting measurements loop
        res = super().exec()
        self.__stop_operations.emit()

        self.__lightweight_workers.clear()
        self.__main_window.wav_player.stop()

        for worker_thread in (self.__lightweight_workers_thread, self.__measurer_thread):
            worker_thread.exit(0)
            worker_thread.wait(10000)  # Wait 10 seconds

        return res

    def __parse_args(self, args: List[str]) -> None:
        parser = argparse.ArgumentParser(description="EyePoint S1")

        parser.add_argument(
            "-a", "-r", "-c", "--reference-oscilloscope",
            dest="reference_oscilloscope", metavar="PATH",
            type=str, nargs=1, help="Oscilloscope A (reference)",
            required=False)
        parser.add_argument(
            "-b", "-t", "--test-oscilloscope",
            dest="test_oscilloscope", metavar="PATH",
            type=str, nargs=1, help="Oscilloscope B (test)",
            required=False)

        parser.add_argument(
            "--window", metavar="TYPE", type=str, nargs="?",
            default=config.window,
            help="Window type (\"pc\" or \"embedded\")")

        parser.add_argument(
            "--screenshot-external-media-path",
            dest="screenshot_external_media_path", metavar="PATH",
            type=str, nargs="?",
            default=config.screenshot_external_media_path,
            help="Path to mounted volumes directory"
                 " (like \"/media/eyepoint/\"). If set, EyePoint will try"
                 " save screenshot to some of mounted media "
                 "(e.g. \"/media/eyepoint/MyDrive/\"")

        parser.add_argument(
            "--screenshot-external-media-only",
            dest="screenshot_external_media_only",
            action="store_true",
            help="Do not offer screenshot save to fallback path")

        parser.add_argument(
            "--screenshot-no-save-as-dialog",
            dest="screenshot_no_save_as_dialog",
            action="store_true",
            help="Don't show \"Save as...\" dialog")

        parser.add_argument(
            "--screenshot-gpio", dest="screenshot_gpio", metavar="NUM",
            type=int, nargs="?", default=config.screenshot_gpio,
            help="Screenshot on GPIO high level")
        parser.add_argument(
            "--screenshot-gpio-inverted", dest="screenshot_gpio_inverted",
            action="store_true", help="Invert screenshot GPIO level")

        parser.add_argument(
            "--freeze-gpio", dest="freeze_gpio", metavar="NUM", type=int,
            nargs="?", default=config.freeze_gpio,
            help="Freeze on GPIO high level")
        parser.add_argument(
            "--freeze-gpio-inverted", dest="freeze_gpio_inverted",
            action="store_true", help="Invert freeze GPIO level")

        parser.add_argument(
            "--hide-cursor", dest="hide_cursor",
            action="store_true", help="Hide mouse cursor")

        parser.add_argument(
            "--debug", action="store_true",
            help="Enable some debug features (like FPS)")
        parser.add_argument(
            "--no-sound", dest="enable_sound_support",
            action="store_false", help="Disable sound")
        parser.add_argument(
            "--no-score", dest="enable_score",
            action="store_false", help="Disable score")
        parser.add_argument(
            "--en", action="store_true", dest="en",
            help="Enable english language")

        parsed_args = parser.parse_args(args[1:])
        self.__english = parsed_args.en
        if parsed_args.reference_oscilloscope:
            self.__reference_oscilloscope_path = parsed_args.reference_oscilloscope[0]
        else:
            self.__reference_oscilloscope_path = ""
        if parsed_args.test_oscilloscope:
            self.__test_oscilloscope_path = parsed_args.test_oscilloscope[0]
        else:
            self.__test_oscilloscope_path = ""

        self.__main_window_type = parsed_args.window.lower()

        self.__screenshot_external_media_path =\
            parsed_args.screenshot_external_media_path
        self.__screenshot_external_media_only = \
            config.screenshot_external_media_only or\
            parsed_args.screenshot_external_media_only
        self.__screenshots_show_save_as_dialog = \
            config.screenshot_show_save_as_dialog and\
            not parsed_args.screenshot_no_save_as_dialog and\
            (parsed_args.screenshot_external_media_path is None)

        self.__screenshot_gpio = parsed_args.screenshot_gpio
        self.__screenshot_gpio_inverted = parsed_args.screenshot_gpio_inverted

        self.__freeze_gpio = parsed_args.freeze_gpio
        self.__freeze_gpio_inverted = parsed_args.freeze_gpio_inverted

        self.__hide_cursor = config.hide_cursor or parsed_args.hide_cursor

        self.__debug = parsed_args.debug
        self.__enable_sound_support = parsed_args.enable_sound_support
        self.__enable_score = parsed_args.enable_score

    def __init_screenshot_manager(self):
        fallback_path = None if self.__screenshot_external_media_only\
            else config.screenshot_fallback_path
        self.__screenshot_file_manager = ScreenshotFileManager(
            directory_name=config.screenshot_directory_name,
            fallback_path=fallback_path,
            external_media_path=self.__screenshot_external_media_path,
            show_save_as_dialog=self.__screenshots_show_save_as_dialog
        )
        self.__screenshot_file_manager.external_media_state_changed.connect(
            self.__on_external_media_state_changed)
        self.__add_workers(*self.__screenshot_file_manager.lightweight_workers)

    def __init_main_window(self):
        screenshots_allowed = not self.__screenshot_external_media_only\
            or self.__screenshot_file_manager.external_media_state\
            == ExternalMediaState.CONNECTED

        enable_sound_support = config.enable_sound_support\
            and self.__enable_sound_support
        self.__main_window = main_window_by_type[self.__main_window_type](
            screenshot_file_manager=self.__screenshot_file_manager,

            score_threshold=float(config.score_threshold),

            enable_sound_support=enable_sound_support,
            enable_sound=bool(config.enable_sound),

            frequency=config.frequency,
            frequency_enabled=config.frequency_enabled,

            voltage=config.voltage,
            voltage_enabled=config.voltage_enabled,

            current=config.current,
            current_enabled=config.current_enabled,

            screenshots_allowed=screenshots_allowed,
            screenshots_show_save_as_dialog=self.__screenshots_show_save_as_dialog,

            debug=self.__debug,

            enable_score=self.__enable_score
        )

        if self.__hide_cursor:
            QApplication.setOverrideCursor(Qt.BlankCursor)

    def __init_gpio_readers(self):
        if self.__screenshot_gpio is not None:
            self.__screenshot_gpio_reader = GpioReader(
                pin=self.__screenshot_gpio,
                frequency=config.screenshot_gpio_check_frequency)
            self.__add_workers(self.__screenshot_gpio_reader)
            self.__screenshot_gpio_reader.state_changed.connect(
                self.__on_screenshot_gpio_state_changed)

        if self.__freeze_gpio is not None:
            self.__freeze_gpio_reader = GpioReader(
                pin=self.__freeze_gpio,
                frequency=config.freeze_gpio_check_frequency)
            self.__add_workers(self.__freeze_gpio_reader)
            self.__freeze_gpio_reader.state_changed.connect(
                self.__on_freeze_gpio_state_changed)

    @QtCore.pyqtSlot(bool)
    def __on_devices_ready(
            self,
            connected: bool
    ) -> None:
        self.__main_window.set_devices_ready(connected)

    @QtCore.pyqtSlot(bool, str, str, str, list)
    def __on_devices_correct(
            self,
            correct: bool,
            name: str,
            version: str,
            firmw: str,
            firms: list
    ) -> None:
        self.__main_window.set_devices_correct(correct, name, version, firmw, firms)

    @QtCore.pyqtSlot(tuple, tuple, Frequency, Voltage, Current)
    def __on_curves_ready(
            self,
            reference_measurement, test_measurement,
            frequency: Frequency, voltage: Voltage, current: Current
    ) -> None:
        self.__measure_required.emit()  # Next iteration...
        self.__main_window.fps = self.__get_fps()
        self.__main_window.set_curves(
            reference_curve=reference_measurement, test_curve=test_measurement,
            frequency=frequency, voltage=voltage, current=current
        )
        return

    def __get_fps(self):
        frame_time = time.monotonic()
        if frame_time - self.__previous_frame_time != 0:
            fps = 1.0 / (frame_time - self.__previous_frame_time)
        else:
            fps = 1000.0
        self.__previous_frame_time = frame_time
        return fps

    @QtCore.pyqtSlot(GpioState)
    def __on_freeze_gpio_state_changed(self, state: GpioState) -> None:
        high = (state == GpioState.HIGH)
        if self.__freeze_gpio_inverted:
            high = not high
        print("Freeze GPIO state changed to {}.".format(
            state.name))
        self.__main_window.on_freeze_gpio(high)

    @QtCore.pyqtSlot(GpioState)
    def __on_screenshot_gpio_state_changed(self, state: GpioState) -> None:
        need_make_screenshot = (state == GpioState.HIGH)
        if self.__screenshot_gpio_inverted:
            need_make_screenshot = not need_make_screenshot

        print(
            "Screenshot GPIO state changed to {}"
            " => {}need make screenshot.".format(
                state.name, "" if need_make_screenshot else "don't "))
        if need_make_screenshot:
            self.__main_window.make_screenshot()

    @QtCore.pyqtSlot(ExternalMediaState)
    def __on_external_media_state_changed(
            self, state: ExternalMediaState) -> None:
        print("External media state is: {}.".format(state.name))

        # here we call self.__main_window.set_screenshots_allowed func
        # because it calls as set parametr of self.__main_window.screenshots_allowed property
        self.__main_window.screenshots_allowed = \
            state == ExternalMediaState.CONNECTED or\
            not self.__screenshot_external_media_only

        # Here we emit File not found, or flash drive connected  behaviour
        if state == ExternalMediaState.DISCONNECTED:
            self.__main_window.on_flash_disconnect()
        if state == ExternalMediaState.CONNECTED:
            self.__main_window.on_flash_connected()
        print(self.__main_window.screenshots_allowed)
