﻿import sys
from enum import Enum
from collections import namedtuple, defaultdict
from functools import partial
import numpy as np

from PyQt5 import QtCore
from PyQt5.QtCore import QObject, QTimer

from ..product.mode import Frequency, Voltage, Current
from scipy import interpolate as itl
from .oscilloscope_common import Oscilloscope, compare_ivc
from .oscilloscope_urpc import OscilloscopeURPC

__all__ = ["OscilloscopeMeasurer"]


class OscilloscopeStatus(Enum):
    SUCCESS = 0
    FAILED = 1


_FrequencyMeasureSettings = namedtuple(
    "_FrequencyInfo", ("discretization_frequency", "probe_signal_frequency"))


_frequency_to_measure_settings = {
    Frequency.FREQUENCY_10HZ: _FrequencyMeasureSettings(1000.0, 10.0),
    Frequency.FREQUENCY_100HZ: _FrequencyMeasureSettings(10000.0, 100.0),
    Frequency.FREQUENCY_1KHZ: _FrequencyMeasureSettings(100000.0, 1000.0),
    Frequency.FREQUENCY_10KHZ: _FrequencyMeasureSettings(1000000.0, 10000.0),
    Frequency.FREQUENCY_100KHZ: _FrequencyMeasureSettings(10000000.0, 100000.0)
}


_voltage_to_max_voltage = {
    Voltage.VOLTAGE_3_3V: 3.3,
    Voltage.VOLTAGE_5V: 5.0,
    Voltage.VOLTAGE_12V: 12.0
}


_CurrentMeasureSettings = namedtuple(
    "_CurrentMeasureSettings", ("max_current", "measure_flag"))


_current_to_measure_settings = {
    Current.CURRENT_250UA: _CurrentMeasureSettings(250.0, 1 | 0x10),
    Current.CURRENT_2_5MA: _CurrentMeasureSettings(250.0, 2 | 0x10),
    Current.CURRENT_25MA: _CurrentMeasureSettings(250.0, 3 | 0x10)
}


_MinVar = namedtuple("_MinVar", ("voltage", "current"))


# Mapping[Tuple[_Voltage, _Current], _MinVar]
# Note: voltage=None, current=None means default values (see compare_ivc()).
_voltage_current_to_min_var = defaultdict(
    lambda: _MinVar(voltage=None, current=None))

# Filling example:
# _voltage_current_to_min_var[(_Voltage.VOLTAGE_12V, _Current.CURRENT_25MA)] =\
#  _MinVar(voltage=0.6, current=0.01)
_voltage_current_to_min_var[(Voltage.VOLTAGE_12V, Current.CURRENT_25MA)] =\
    _MinVar(voltage=0.6, current=0.008)
_voltage_current_to_min_var[(Voltage.VOLTAGE_5V, Current.CURRENT_25MA)] =\
    _MinVar(voltage=0.6, current=0.008)
_voltage_current_to_min_var[(Voltage.VOLTAGE_3_3V, Current.CURRENT_25MA)] =\
    _MinVar(voltage=0.3, current=0.008)
_voltage_current_to_min_var[(Voltage.VOLTAGE_12V, Current.CURRENT_2_5MA)] =\
    _MinVar(voltage=0.6, current=0.0005)
_voltage_current_to_min_var[(Voltage.VOLTAGE_5V, Current.CURRENT_2_5MA)] =\
    _MinVar(voltage=0.3, current=0.0005)
_voltage_current_to_min_var[(Voltage.VOLTAGE_3_3V, Current.CURRENT_2_5MA)] =\
    _MinVar(voltage=0.3, current=0.0005)
_voltage_current_to_min_var[(Voltage.VOLTAGE_12V, Current.CURRENT_250UA)] =\
    _MinVar(voltage=0.6, current=0.00005)
_voltage_current_to_min_var[(Voltage.VOLTAGE_5V, Current.CURRENT_250UA)] =\
    _MinVar(voltage=0.3, current=0.00005)
_voltage_current_to_min_var[(Voltage.VOLTAGE_3_3V, Current.CURRENT_250UA)] =\
    _MinVar(voltage=0.3, current=0.00005)


def compare_ivc_ex(reference_curve, test_curve, voltage, current):
    min_var_voltage, min_var_current = _voltage_current_to_min_var[
        (voltage, current)]
    return compare_ivc(
        reference_curve,
        test_curve,
        min_var_v=min_var_voltage,
        min_var_c=min_var_current)


class OscilloscopeMeasurer(QObject):
    devices_ready = QtCore.pyqtSignal(
        bool)
    devices_not_correct = QtCore.pyqtSignal(
        bool, str, str, str, list)
    curves_ready = QtCore.pyqtSignal(
        tuple, tuple, Frequency, Voltage, Current)

    def __init__(
            self,
            reference_oscilloscope_path: str,
            test_oscilloscope_path: str,

            frequency: Frequency = Frequency.FREQUENCY_10HZ,
            voltage: Voltage = Voltage.VOLTAGE_3_3V,
            current: Current = Current.CURRENT_250UA,

            parent: QObject = None
    ) -> None:
        super().__init__(parent=parent)

        self.__reference_oscilloscope = self._get_oscilloscope(
            reference_oscilloscope_path
        )

        self.__test_oscilloscope = self._get_oscilloscope(
            test_oscilloscope_path
        )

        self.__oscilloscope_settings = None

        self.__frequency = frequency
        self.__old_frequency = None
        self.__frozen_frequency = None

        self.__voltage = voltage
        self.__old_voltage = None
        self.__frozen_voltage = None

        self.__current = current
        self.__old_current = None
        self.__frozen_current = None

        self.__freeze_reference_curve = False
        self.__freeze_test_curve = False

        self.__reference_curve = None
        self.__test_curve = None

    def _is_virtual(self, device: str) -> bool:
        return device.lower() == "virtual"

    def _get_oscilloscope(self, device: str) -> Oscilloscope:
        if self._is_virtual(device):
            return Oscilloscope()
        else:
            return OscilloscopeURPC(device_name=device)

    def __get_devices(self):
        devs = (
            self.__reference_oscilloscope,
            self.__test_oscilloscope
        )
        do_start = False
        self.status = 0
        Name = " "
        version = " "
        firmw = " "
        firms = []
        _devs = []
        for dev in devs:
            try:
                if not dev.is_opened:
                    ind, Name, version, firmw, firms = dev.open_device()
                    self.status += ind
                    if ind > 0:
                        _devs.append(dev)
                        self.__apply_oscilloscope_settings(dev)
                        do_start = True
                        self.__apply_oscilloscope_settings(dev)
                else:
                    self.status += 1
                    _devs.append(dev)
            except RuntimeError:
                _devs.append(None)
        if _devs == [None, None]:
            self.__close_devices()
            return None
        elif _devs[0] is None:
            _devs.reverse()
        if do_start:
            self.start()
        self.devices_not_correct.emit(not (self.status > 0), Name, version, firmw, firms)
        self.devices_ready.emit(True)
        return tuple(_devs) if len(_devs) == 2 else None

    def __close_devices(self):
        self.__reference_oscilloscope.close_device()
        self.__test_oscilloscope.close_device()
        self.devices_ready.emit(False)

    @QtCore.pyqtSlot(Frequency)
    def set_frequency(self, frequency: Frequency) -> None:
        self.__frequency = frequency

    @QtCore.pyqtSlot(Voltage)
    def set_voltage(self, voltage: Voltage) -> None:
        self.__voltage = voltage

    @QtCore.pyqtSlot(Current)
    def set_current(self, current: Current) -> None:
        self.__current = current

    @QtCore.pyqtSlot(bool)
    def set_freeze_reference_curve(self, freeze_reference_curve: bool) -> None:
        self.__freeze_reference_curve = freeze_reference_curve
        self.__save_frozen_parameters()

    @QtCore.pyqtSlot(bool)
    def set_freeze_test_curve(self, freeze_test_curve: bool) -> None:
        self.__freeze_test_curve = freeze_test_curve
        self.__save_frozen_parameters()

    def __save_frozen_parameters(self):
        self.__frozen_frequency = self.__frequency
        self.__frozen_voltage = self.__voltage
        self.__frozen_current = self.__current

    @QtCore.pyqtSlot()
    def start(self) -> None:
        self.calibrate_oscilloscopes()
        self.__update_measure_settings()

    @QtCore.pyqtSlot()
    def stop(self) -> None:
        pass

    # TODO Move methods below to a separate curve processor class
    def __extract_single_period(self, curve, settings):
        """
        If the curve has multiple periods,
        this function extracts first single period
        :param curve: list or tuple of two arrays with voltages and currents
        :param settings: Measurement settings
        :return: curve with extracted first period in format same to the curve, but with fewer points
        """
        period_n_points = int(settings.desc_frequency // settings.probe_signal_frequency)
        curve_cutted_0 = curve[0][0:period_n_points]
        curve_cutted_1 = curve[1][0:period_n_points]
        return np.array((curve_cutted_0, curve_cutted_1), dtype=np.float32)

    def __interpolate_curve(self, curve, final_num_points):
        """
        Effectively increase resolution by curve interpolation
        :param curve: list or tuple of two arrays with voltages and currents
        :param final_num_points: Number points in final interpolated curve
        :return: interpolated curve in format same to the curve
        """
        curve = np.array(
            (curve[0, 1:], curve[1, 1:]), dtype=np.float32
        )

        curve_len = curve.shape[1]
        raw_t = np.linspace(0, curve_len - 1, curve_len, dtype=np.float32)

        interp_function_0 = itl.interp1d(raw_t, curve[0], kind="cubic")
        interp_function_1 = itl.interp1d(raw_t, curve[1], kind="cubic")
        interpolated_t = np.linspace(0, curve_len - 1, final_num_points, dtype=np.float32)
        interp0 = interp_function_0(interpolated_t)
        interp1 = interp_function_1(interpolated_t)

        return np.array((interp0, interp1), dtype=np.float32)

    def __smooth_curve(self, curve, kernel_size):
        """
        Remove noise by averaging
        :param curve: list or tuple of two arrays with voltages and currents
        :param kernel_size: size of averaging kernel. Should be odd.
        :return: averaged curve in format same to the curve
        """

        if kernel_size % 2 == 0:
            raise ValueError("kernel_size should be odd")

        kernel = np.ones(kernel_size) / kernel_size

        smoothed_curve = []
        for line in curve:
            line = np.array(line)
            line = np.concatenate((line[-(kernel_size - 1) // 2:], line, line[:(kernel_size - 1) // 2]))
            line = np.convolve(line, kernel, mode="valid")
            smoothed_curve.append(line)

        return smoothed_curve

    def __eliminate_parasitic_serial_resistance(self, curve, resistance):
        """
        Remove parasitic probe resistance.
        Model resistance is connected in serial:
        I real = I
        U real = U - I * R
        :param curve: list or tuple of two arrays with voltages and currents
        :param resistance: Parasitic resistance in Ohms.
        :return: processed curve in format same to the curve
        """
        voltages = np.array(curve[0])
        currents = np.array(curve[1])

        voltages = voltages - currents * resistance

        return (voltages, currents)

    def __measure(self, osc):
        s, curve = osc.measure()

        # If the curve has multiple periods (but with lower resolution),
        # we cut only one and add more points by interpolations
        if s.desc_frequency // s.probe_signal_frequency < s.n_points:
            curve = self.__extract_single_period(curve, s)
            curve = self.__interpolate_curve(curve, s.n_points)

        # Remove noise by averaging
        curve = self.__smooth_curve(curve, self.SMOOTHING_KERNEL_SIZE)

        # Remove parasitic currents
        # curve = self.__eliminate_parasitic_serial_resistance(curve, self.PARASITIC_SERIAL_RESISTANCE)

        return s, curve

    @QtCore.pyqtSlot()
    def async_measure(self) -> None:
        error = False

        devs = self.__get_devices()
        if devs is None:
            error = True
        else:
            ref_osc, test_osc = devs
            self.__update_measure_settings()

        measured = False
        if not error and (not self.__freeze_reference_curve or self.__reference_curve is None) and ref_osc is not None:
            try:
                self.__reference_curve = self.__measure(ref_osc)
                measured = True
            except Exception:
                error = True
        else:
            self.__reference_curve = (None, None)

        if not error and (not self.__freeze_test_curve or self.__test_curve is None) and test_osc is not None:
            try:
                self.__test_curve = self.__measure(test_osc)
                measured = True
            except Exception:
                error = True
        else:
            self.__test_curve = (None, None)

        if error:
            self.__close_devices()
            QTimer.singleShot(100, self.async_measure)
            return

        if measured:
            self.__report_curves()
            # QTimer.singleShot(100, partial(self.__report_curves, False))
        else:
            # Report with delay to reduce CPU load
            QTimer.singleShot(100, partial(self.__report_curves, True))

    def __report_curves(self, frozen: bool = False) -> None:
        if frozen:
            self.curves_ready.emit(
                self.__reference_curve, self.__test_curve,
                self.__frozen_frequency, self.__frozen_voltage,
                self.__frozen_current)
        else:
            self.curves_ready.emit(
                self.__reference_curve, self.__test_curve,
                self.__frequency, self.__voltage, self.__current)

    @QtCore.pyqtSlot()
    def calibrate_oscilloscopes(self) -> None:
        devs = self.__get_devices()
        if devs is None:
            return
        ref_osc, test_osc = devs

        try:
            # TODO: Improve workaround with config file.
            # if isinstance(ref_osc, OscilloscopeURPC):
            #     ref_osc.manual_calibrate_7()
            # else:
            if ref_osc is not None:
                ref_osc.full_calibrate()
            else:
                pass
        except Exception as e:
            print(
                "When calibrating reference oscilloscope caught: {}."
                .format(e),
                file=sys.stderr)
            self.__close_devices()
            return

        try:
            # TODO: Improve workaround with config file.
            # if isinstance(test_osc, OscilloscopeURPC):
            #     test_osc.manual_calibrate_5()
            # else:
            if test_osc is not None:
                test_osc.full_calibrate()
            else:
                pass
        except Exception as e:
            print("When calibrating oscilloscope caught: {}.".format(e),
                  file=sys.stderr)
            self.__close_devices()
            return

    # Constants
    CURVE_LENGTH = 100  # Number points in final curve
    SMOOTHING_KERNEL_SIZE = 5
    PARASITIC_SERIAL_RESISTANCE = 80  # 80 Ohms probe resistance, connected in serial

    def __update_measure_settings(self) -> None:
        if self.__frequency != self.__old_frequency \
                or self.__voltage != self.__old_voltage \
                or self.__current != self.__old_current:
            self.__old_frequency = self.__frequency
            self.__old_voltage = self.__voltage
            self.__old_current = self.__current

            fms = _frequency_to_measure_settings[self.__frequency]
            cms = _current_to_measure_settings[self.__current]

            d = dict(
                desc_frequency=fms.discretization_frequency,
                probe_signal_frequency=fms.probe_signal_frequency,
                max_voltage=_voltage_to_max_voltage[self.__voltage],
                # TODO: Uncomment and fix values.
                # max_current=cms.max_current,
                n_points=self.CURVE_LENGTH,
                n_charge_points=0,
                flags=cms.measure_flag
            )
            # 2 MSPS - is a maximal IVM discretization frequency
            # So 100 kHz meaurements will have only 20 points instead of 100
            if self.__frequency == Frequency.FREQUENCY_100KHZ:
                d.update(dict(
                    desc_frequency=2000000.,
                    probe_signal_frequency=100000.,
                    n_points=self.CURVE_LENGTH,
                    n_charge_points=0
                    # max_current=25
                ))
            self.__set_measure_settings(**d)

    def __set_measure_settings(self, **settings) -> None:
        self.__oscilloscope_settings = dict(settings)
        devs = self.__get_devices()
        if devs is None:
            return
        self.__apply_oscilloscope_settings(*devs)

    def __apply_oscilloscope_settings(self, *devs):
        if self.__oscilloscope_settings is None:
            return

        for dev in devs:
            try:
                if dev is not None:
                    dev.set_settings(**self.__oscilloscope_settings)
                else:
                    pass
            except Exception as e:
                print(
                    "When setting oscilloscope parameters caught: {}."
                    .format(e),
                    file=sys.stderr)
                self.__close_devices()
                return
